@@ -48,6 +48,27 @@ struct ipc_auth_ugp {
 	pid_t pid;
 };
 
+struct ipc_auth_data {
+	int32_t sock;
+	struct qb_ipcs_service *s;
+	union {
+		struct qb_ipc_connection_request req;
+		struct qb_ipc_connection_response res;
+	} msg;
+	struct msghdr msg_recv;
+	struct iovec iov_recv;
+	struct ipc_auth_ugp ugp;
+
+	size_t processed;
+	size_t len;
+
+#ifdef SO_PASSCRED
+	char *cmsg_cred;
+#endif
+
+};
+
+
 static int32_t qb_ipcs_us_connection_acceptor(int fd, int revent, void *data);
 
 ssize_t
@@ -83,20 +104,21 @@ retry_send:
 }
 
 static ssize_t
-qb_ipc_us_recv_msghdr(int32_t s, struct msghdr *hdr, char *msg, size_t len)
+qb_ipc_us_recv_msghdr(struct ipc_auth_data *data)
 {
+	char *msg = (char *) &data->msg;
 	int32_t result;
-	int32_t processed = 0;
 
 	qb_sigpipe_ctl(QB_SIGPIPE_IGNORE);
 
 retry_recv:
-	hdr->msg_iov->iov_base = &msg[processed];
-	hdr->msg_iov->iov_len = len - processed;
+	data->msg_recv.msg_iov->iov_base = &msg[data->processed];
+	data->msg_recv.msg_iov->iov_len = data->len - data->processed;
 
-	result = recvmsg(s, hdr, MSG_NOSIGNAL | MSG_WAITALL);
+	result = recvmsg(data->sock, &data->msg_recv, MSG_NOSIGNAL | MSG_WAITALL);
 	if (result == -1 && errno == EAGAIN) {
-		goto retry_recv;
+		qb_sigpipe_ctl(QB_SIGPIPE_DEFAULT);
+		return -EAGAIN;
 	}
 	if (result == -1) {
 		qb_sigpipe_ctl(QB_SIGPIPE_DEFAULT);
@@ -105,18 +127,18 @@ retry_recv:
 	if (result == 0) {
 		qb_sigpipe_ctl(QB_SIGPIPE_DEFAULT);
 		qb_util_log(LOG_DEBUG,
-			    "recv(fd %d) got 0 bytes assuming ENOTCONN", s);
+			    "recv(fd %d) got 0 bytes assuming ENOTCONN", data->sock);
 		return -ENOTCONN;
 	}
 
-	processed += result;
-	if (processed != len) {
+	data->processed += result;
+	if (data->processed != data->len) {
 		goto retry_recv;
 	}
 	qb_sigpipe_ctl(QB_SIGPIPE_DEFAULT);
-	assert(processed == len);
+	assert(data->processed == data->len);
 
-	return processed;
+	return data->processed;
 }
 
 int32_t
@@ -266,7 +288,7 @@ qb_ipcc_stream_sock_connect(const char *socket_name, int32_t * sock_pt)
 #if defined(QB_LINUX) || defined(QB_CYGWIN)
 	snprintf(address.sun_path + 1, UNIX_PATH_MAX - 1, "%s", socket_name);
 #else
-	snprintf(address.sun_path, UNIX_PATH_MAX, "%s/%s", SOCKETDIR,
+	snprintf(address.sun_path, sizeof(address.sun_path), "%s/%s", SOCKETDIR,
 		 socket_name);
 #endif
 	if (connect(request_fd, (struct sockaddr *)&address,
@@ -292,12 +314,138 @@ qb_ipcc_us_sock_close(int32_t sock)
 	close(sock);
 }
 
+static int32_t
+qb_ipc_auth_creds(struct ipc_auth_data *data)
+{
+	int32_t res = 0;
+
+	/*
+	 * currently support getpeerucred, getpeereid, and SO_PASSCRED credential
+	 * retrieval mechanisms for various Platforms
+	 */
+#ifdef HAVE_GETPEERUCRED
+	/*
+	 * Solaris and some BSD systems
+	 */
+	{
+		ucred_t *uc = NULL;
+
+		if (getpeerucred(data->sock, &uc) == 0) {
+			res = 0;
+			data->ugp.uid = ucred_geteuid(uc);
+			data->ugp.gid = ucred_getegid(uc);
+			data->ugp.pid = ucred_getpid(uc);
+			ucred_free(uc);
+		} else {
+			res = -errno;
+		}
+	}
+#elif defined(HAVE_GETPEEREID)
+	/*
+	* Usually MacOSX systems
+	*/
+	{
+		/*
+		* TODO get the peer's pid.
+		* c->pid = ?;
+		*/
+		if (getpeereid(data->sock, &data->ugp.uid, &data->ugp.gid) == 0) {
+			res = 0;
+		} else {
+			res = -errno;
+		}
+	}
+
+#elif defined(SO_PASSCRED)
+	/*
+	* Usually Linux systems
+	*/
+	{
+		struct ucred cred;
+		struct cmsghdr *cmsg;
+
+		res = -EINVAL;
+		for (cmsg = CMSG_FIRSTHDR(&data->msg_recv); cmsg != NULL;
+			cmsg = CMSG_NXTHDR(&data->msg_recv, cmsg)) {
+			if (cmsg->cmsg_type != SCM_CREDENTIALS)
+				continue;
+
+			memcpy(&cred, CMSG_DATA(cmsg), sizeof(struct ucred));
+			res = 0;
+			data->ugp.pid = cred.pid;
+			data->ugp.uid = cred.uid;
+			data->ugp.gid = cred.gid;
+			break;
+		}
+	}
+#else /* no credentials */
+	data->ugp.pid = 0;
+	data->ugp.uid = 0;
+	data->ugp.gid = 0;
+	res = -ENOTSUP;
+#endif /* no credentials */
+
+	return res;
+}
+
+static void
+destroy_ipc_auth_data(struct ipc_auth_data *data)
+{
+	if (data->s) {
+		qb_ipcs_unref(data->s);
+	}
+
+#ifdef SO_PASSCRED
+	free(data->cmsg_cred);
+#endif
+	free(data);
+}
+
+static struct ipc_auth_data *
+init_ipc_auth_data(int sock, size_t len)
+{
+	struct ipc_auth_data *data = calloc(1, sizeof(struct ipc_auth_data));
+
+	if (data == NULL) {
+		return NULL;
+	}
+
+	data->msg_recv.msg_iov = &data->iov_recv;
+	data->msg_recv.msg_iovlen = 1;
+	data->msg_recv.msg_name = 0;
+	data->msg_recv.msg_namelen = 0;
+
+#ifdef SO_PASSCRED
+	data->cmsg_cred = calloc(1, CMSG_SPACE(sizeof(struct ucred)));
+	if (data->cmsg_cred == NULL) {
+		destroy_ipc_auth_data(data);
+		return NULL;
+	}
+	data->msg_recv.msg_control = (void *)data->cmsg_cred;
+	data->msg_recv.msg_controllen = CMSG_SPACE(sizeof(struct ucred));
+#endif
+#ifdef QB_SOLARIS
+	data->msg_recv.msg_accrights = 0;
+	data->msg_recv.msg_accrightslen = 0;
+#else
+	data->msg_recv.msg_flags = 0;
+#endif /* QB_SOLARIS */
+
+	data->len = len;
+	data->iov_recv.iov_base = &data->msg;
+	data->iov_recv.iov_len = data->len;
+	data->sock = sock;
+
+	return data;
+}
+
 int32_t
 qb_ipcc_us_setup_connect(struct qb_ipcc_connection *c,
 			 struct qb_ipc_connection_response *r)
 {
 	int32_t res;
 	struct qb_ipc_connection_request request;
+	struct ipc_auth_data *data;
 #ifdef QB_LINUX
 	int off = 0;
 	int on = 1;
@@ -321,22 +469,33 @@ qb_ipcc_us_setup_connect(struct qb_ipcc_connection *c,
 		qb_ipcc_us_sock_close(c->setup.u.us.sock);
 		return res;
 	}
+
+	data = init_ipc_auth_data(c->setup.u.us.sock, sizeof(struct qb_ipc_connection_response));
+	if (data == NULL) {
+		qb_ipcc_us_sock_close(c->setup.u.us.sock);
+		return -ENOMEM;
+	}
+
+	qb_ipc_us_ready(&c->setup, NULL, -1, POLLIN);
+	res = qb_ipc_us_recv_msghdr(data);
+
 #ifdef QB_LINUX
 	setsockopt(c->setup.u.us.sock, SOL_SOCKET, SO_PASSCRED, &off,
 		   sizeof(off));
 #endif
 
-	res =
-	    qb_ipc_us_recv(&c->setup, r,
-			   sizeof(struct qb_ipc_connection_response), -1);
-	if (res < 0) {
+	if (res != data->len) {
+		destroy_ipc_auth_data(data);
 		return res;
 	}
 
-	if (r->hdr.error != 0) {
-		return r->hdr.error;
-	}
-	return 0;
+	memcpy(r, &data->msg.res, sizeof(struct qb_ipc_connection_response));
+
+	qb_ipc_auth_creds(data);
+	c->egid = data->ugp.gid;
+
+	destroy_ipc_auth_data(data);
+	return r->hdr.error;
 }
 
 /*
@@ -349,6 +508,9 @@ qb_ipcs_us_publish(struct qb_ipcs_service * s)
 {
 	struct sockaddr_un un_addr;
 	int32_t res;
+#ifdef SO_PASSCRED
+	int on = 1;
+#endif
 
 	/*
 	 * Create socket for IPC clients, name socket, listen for connections
@@ -385,7 +547,7 @@ qb_ipcs_us_publish(struct qb_ipcs_service * s)
 				    SOCKETDIR);
 			goto error_close;
 		}
-		snprintf(un_addr.sun_path, UNIX_PATH_MAX, "%s/%s", SOCKETDIR,
+		snprintf(un_addr.sun_path, sizeof(un_addr.sun_path), "%s/%s", SOCKETDIR,
 			 s->name);
 		unlink(un_addr.sun_path);
 	}
@@ -406,6 +568,9 @@ qb_ipcs_us_publish(struct qb_ipcs_service * s)
 	 */
 #if !defined(QB_LINUX) && !defined(QB_CYGWIN)
 	res = chmod(un_addr.sun_path, S_IRWXU | S_IRWXG | S_IRWXO);
+#endif
+#ifdef SO_PASSCRED
+	setsockopt(s->server_sock, SOL_SOCKET, SO_PASSCRED, &on, sizeof(on));
 #endif
 	if (listen(s->server_sock, SERVER_BACKLOG) == -1) {
 		qb_util_perror(LOG_ERR, "socket listen failed");
@@ -425,8 +590,10 @@ int32_t
 qb_ipcs_us_withdraw(struct qb_ipcs_service * s)
 {
 	qb_util_log(LOG_INFO, "withdrawing server sockets");
+	(void)s->poll_fns.dispatch_del(s->server_sock);
 	shutdown(s->server_sock, SHUT_RDWR);
 	close(s->server_sock);
+	s->server_sock = -1;
 	return 0;
 }
 
@@ -440,6 +607,7 @@ handle_new_connection(struct qb_ipcs_service *s,
 	struct qb_ipc_connection_request *req = msg;
 	int32_t res = auth_result;
 	int32_t res2 = 0;
+	uint32_t max_buffer_size = QB_MAX(req->max_msg_size, s->max_buffer_size);
 	struct qb_ipc_connection_response response;
 
 	c = qb_ipcs_connection_alloc(s);
@@ -448,16 +616,16 @@ handle_new_connection(struct qb_ipcs_service *s,
 		return -ENOMEM;
 	}
 
-	c->receive_buf = calloc(1, req->max_msg_size);
+	c->receive_buf = calloc(1, max_buffer_size);
 	if (c->receive_buf == NULL) {
 		free(c);
 		qb_ipcc_us_sock_close(sock);
 		return -ENOMEM;
 	}
 	c->setup.u.us.sock = sock;
-	c->request.max_msg_size = req->max_msg_size;
-	c->response.max_msg_size = req->max_msg_size;
-	c->event.max_msg_size = req->max_msg_size;
+	c->request.max_msg_size = max_buffer_size;
+	c->response.max_msg_size = max_buffer_size;
+	c->event.max_msg_size = max_buffer_size;
 	c->pid = ugp->pid;
 	c->auth.uid = c->euid = ugp->uid;
 	c->auth.gid = c->egid = ugp->gid;
@@ -519,133 +687,121 @@ send_response:
 		if (res == -EACCES) {
 			qb_util_log(LOG_ERR, "Invalid IPC credentials (%s).",
 				    c->description);
+		} else if (res == -EAGAIN) {
+			qb_util_log(LOG_WARNING, "Denied connection, is not ready (%s)",
+				    c->description);
 		} else {
 			errno = -res;
 			qb_util_perror(LOG_ERR,
 				       "Error in connection setup (%s)",
 				       c->description);
 		}
-		qb_ipcs_disconnect(c);
+
+		if (c->state == QB_IPCS_CONNECTION_INACTIVE) {
+			/* This removes the initial alloc ref */
+			qb_ipcs_connection_unref(c);
+		        qb_ipcc_us_sock_close(sock);
+		} else {
+			qb_ipcs_disconnect(c);
+		}
 	}
 	return res;
 }
 
 static int32_t
-qb_ipcs_uc_recv_and_auth(int32_t sock, void *msg, size_t len,
-			 struct ipc_auth_ugp *ugp)
+process_auth(int32_t fd, int32_t revents, void *d)
 {
+	struct ipc_auth_data *data = (struct ipc_auth_data *) d;
+
 	int32_t res = 0;
-	struct msghdr msg_recv;
-	struct iovec iov_recv;
-
 #ifdef SO_PASSCRED
-	char cmsg_cred[CMSG_SPACE(sizeof(struct ucred))];
 	int off = 0;
-	int on = 1;
-#endif
-	msg_recv.msg_iov = &iov_recv;
-	msg_recv.msg_iovlen = 1;
-	msg_recv.msg_name = 0;
-	msg_recv.msg_namelen = 0;
-#ifdef SO_PASSCRED
-	msg_recv.msg_control = (void *)cmsg_cred;
-	msg_recv.msg_controllen = sizeof(cmsg_cred);
-#endif
-#ifdef QB_SOLARIS
-	msg_recv.msg_accrights = 0;
-	msg_recv.msg_accrightslen = 0;
-#else
-	msg_recv.msg_flags = 0;
-#endif /* QB_SOLARIS */
-
-	iov_recv.iov_base = msg;
-	iov_recv.iov_len = len;
-#ifdef SO_PASSCRED
-	setsockopt(sock, SOL_SOCKET, SO_PASSCRED, &on, sizeof(on));
 #endif
 
-	res = qb_ipc_us_recv_msghdr(sock, &msg_recv, msg, len);
-	if (res < 0) {
+	if (data->s->server_sock == -1) {
+		qb_util_log(LOG_DEBUG, "Closing fd (%d) for server shutdown", fd);
+		res = -ESHUTDOWN;
 		goto cleanup_and_return;
 	}
-	if (res != len) {
+
+	if (revents & POLLNVAL) {
+		qb_util_log(LOG_DEBUG, "NVAL conn fd (%d)", fd);
+		res = -EINVAL;
+		goto cleanup_and_return;
+	}
+	if (revents & POLLHUP) {
+		qb_util_log(LOG_DEBUG, "HUP conn fd (%d)", fd);
+		res = -ESHUTDOWN;
+		goto cleanup_and_return;
+	}
+	if ((revents & POLLIN) == 0) {
+		return 0;
+	}
+
+	res = qb_ipc_us_recv_msghdr(data);
+	if (res == -EAGAIN) {
+		/* yield to mainloop, Let mainloop call us again */
+		return 0;
+	}
+
+	if (res != data->len) {
 		res = -EIO;
 		goto cleanup_and_return;
 	}
 
-	/*
-	 * currently support getpeerucred, getpeereid, and SO_PASSCRED credential
-	 * retrieval mechanisms for various Platforms
-	 */
-#ifdef HAVE_GETPEERUCRED
-	/*
-	 * Solaris and some BSD systems
-	 */
-	{
-		ucred_t *uc = NULL;
-
-		if (getpeerucred(sock, &uc) == 0) {
-			res = 0;
-			ugp->uid = ucred_geteuid(uc);
-			ugp->gid = ucred_getegid(uc);
-			ugp->pid = ucred_getpid(uc);
-			ucred_free(uc);
-		} else {
-			res = -errno;
-		}
-	}
-#elif HAVE_GETPEEREID
-	/*
-	 * Usually MacOSX systems
-	 */
-	{
-		/*
-		 * TODO get the peer's pid.
-		 * c->pid = ?;
-		 */
-		if (getpeereid(sock, &ugp->uid, &ugp->gid) == 0) {
-			res = 0;
-		} else {
-			res = -errno;
-		}
-	}
-
-#elif SO_PASSCRED
-	/*
-	 * Usually Linux systems
-	 */
-	{
-		struct ucred cred;
-		struct cmsghdr *cmsg;
-
-		res = -EINVAL;
-		for (cmsg = CMSG_FIRSTHDR(&msg_recv); cmsg != NULL;
-		     cmsg = CMSG_NXTHDR(&msg_recv, cmsg)) {
-			if (cmsg->cmsg_type != SCM_CREDENTIALS)
-				continue;
-
-			memcpy(&cred, CMSG_DATA(cmsg), sizeof(struct ucred));
-			res = 0;
-			ugp->pid = cred.pid;
-			ugp->uid = cred.uid;
-			ugp->gid = cred.gid;
-			break;
-		}
-	}
-#else /* no credentials */
-	ugp->pid = 0;
-	ugp->uid = 0;
-	ugp->gid = 0;
-	res = -ENOTSUP;
-#endif /* no credentials */
+	res = qb_ipc_auth_creds(data);
 
 cleanup_and_return:
-
 #ifdef SO_PASSCRED
-	setsockopt(sock, SOL_SOCKET, SO_PASSCRED, &off, sizeof(off));
+	setsockopt(data->sock, SOL_SOCKET, SO_PASSCRED, &off, sizeof(off));
 #endif
 
-	return res;
+	(void)data->s->poll_fns.dispatch_del(data->sock);
+
+	if (res < 0) {
+		close(data->sock);
+	} else if (data->msg.req.hdr.id == QB_IPC_MSG_AUTHENTICATE) {
+		(void)handle_new_connection(data->s, res, data->sock, &data->msg, data->len, &data->ugp);
+	} else {
+		close(data->sock);
+	}
+	destroy_ipc_auth_data(data);
+
+	return 1;
+}
+
+static void
+qb_ipcs_uc_recv_and_auth(int32_t sock, struct qb_ipcs_service *s)
+{
+	int res = 0;
+	struct ipc_auth_data *data = NULL;
+#ifdef SO_PASSCRED
+	int on = 1;
+#endif
+
+	data = init_ipc_auth_data(sock, sizeof(struct qb_ipc_connection_request));
+	if (data == NULL) {
+		close(sock);
+		/* -ENOMEM */
+		return;
+	}
+
+	data->s = s;
+	qb_ipcs_ref(data->s);
+
+#ifdef SO_PASSCRED
+	setsockopt(sock, SOL_SOCKET, SO_PASSCRED, &on, sizeof(on));
+#endif
+
+	res = s->poll_fns.dispatch_add(QB_LOOP_MED,
+					data->sock,
+					POLLIN | POLLPRI | POLLNVAL,
+					data, process_auth);
+	if (res < 0) {
+		qb_util_log(LOG_DEBUG, "Failed to process AUTH for fd (%d)", data->sock);
+		close(sock);
+		destroy_ipc_auth_data(data);
+	}
 }
 
 static int32_t
@@ -655,8 +811,6 @@ qb_ipcs_us_connection_acceptor(int fd, int revent, void *data)
 	int32_t new_fd;
 	struct qb_ipcs_service *s = (struct qb_ipcs_service *)data;
 	int32_t res;
-	struct qb_ipc_connection_request setup_msg;
-	struct ipc_auth_ugp ugp;
 	socklen_t addrlen = sizeof(struct sockaddr_un);
 
 	if (revent & (POLLNVAL | POLLHUP | POLLERR)) {
@@ -696,22 +850,6 @@ retry_accept:
 		return 0;
 	}
 
-	res = qb_ipcs_uc_recv_and_auth(new_fd, &setup_msg, sizeof(setup_msg),
-				       &ugp);
-	if (res < 0) {
-		close(new_fd);
-		/* This is an error, but -1 would indicate disconnect
-		 * from the poll loop
-		 */
-		return 0;
-	}
-
-	if (setup_msg.hdr.id == QB_IPC_MSG_AUTHENTICATE) {
-		(void)handle_new_connection(s, res, new_fd, &setup_msg,
-					    sizeof(setup_msg), &ugp);
-	} else {
-		close(new_fd);
-	}
-
+	qb_ipcs_uc_recv_and_auth(new_fd, s);
 	return 0;
 }
